
typedef Vec<void*, 16> PointerList ;

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//    The `RMSchedule' class.                                          //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class RMSchedule {
public :
  void gather() ;
//  void scatter() ;
//  void combine(void (*op)(void*, void*)) ;

protected :
  RMSchedule(void* _locBase, const int _remOff, void** _remBases,
             const int _elementLen) :
		    locBase(_locBase), remOff(_remOff), remBases(_remBases),
		    elementLen(_elementLen) {
    MPI_Comm_rank(MPI_COMM_WORLD, &locId) ;
    MPI_Comm_size(MPI_COMM_WORLD, &nProcs) ;

    locPtrs = new PointerList [nProcs] ;
    remPtrs = new PointerList [nProcs] ;
  }

  void transReq(int locOffset, int remId, int remOffset) {

    // Add a transfer request to the appropriate `locIdxs', `remIdxs' vectors.

    (locPtrs [remId]).add((char*) locBase + elementLen * locOffset) ;
    (remPtrs [remId]).add((char*) remBases [remId] + 
                          elementLen * (remOff + remOffset)) ;
  }

  ~RMSchedule() {
    delete [] remPtrs ;
    delete [] locPtrs ;
  }

private :
  int nProcs, locId ;

  int elementLen ;
  void* locBase ;
  int remOff ;
  void** remBases ;

  PointerList* locPtrs ;
  PointerList* remPtrs ;
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//    The `CollRMSchedule' class.  Collective version---adds           //
//                                 synchronization.                    //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class CollRMSchedule : public RMSchedule {
public :
  CollRMSchedule(void* _locBase, const int _remOff, void** _remBases,
                 const int _elementLen) :
          RMSchedule(_locBase, _remOff, _remBases, _elementLen) {}

  void build(TreeSchedule* _spanTree) {spanTree = _spanTree ; }

  void gather() {
    spanTree->synch() ;
    RMSchedule :: gather() ;
    spanTree->synch() ;
  }

//  void scatter() ;
//  void combine(void (*op)(void*, void*)) ;

private :
  TreeSchedule* spanTree ;
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//    Adding block transfers to queues.                                //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //


#define MBLKT_REM 0    // Remote processor id.
#define MBLKT_SIZ 1    // Number of bytes in block
#define MBLKT_NST 2    // Number of stride levels.
#define MBLKT_CTG 3    // Length of contiguous blocks in bytes
#define MBLKT_SHP 4    // First of `NST' extents (followed by `NST' strides
                       // for 1st block, then `NST' strides for 2nd block).

// NB: Unit of strides is one byte (*not* one array element).
// NB: If these constants are changed, `transReq' below must be changed.


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//    The `BlockRMSchedule' class.                                     //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class BlockRMSchedule {
public :
  BlockRMSchedule(void* _locBase, const int _remOff, void** _remBases,
                  const int _elementLen, const int _rank) :
               locBase(_locBase), remOff(_remOff), remBases(_remBases),
               rank(_rank), elementLen(_elementLen),
               blockLen(MBLKT_SHP + 3 * _rank) {
    numBlk = 0 ;
  }

  void gather() ;
  void scatter() ;

  void transReq(int locOffset, int* locStrs,
		int remOffset, int* remStrs,
                int* exts, int numStrLevels, int contigLen,
                int vol, int remId) {

    // Add a transfer request to the appropriate lists.

    locPtrs.add((char*) locBase + elementLen * locOffset) ;
    remPtrs.add((char*) remBases [remId] + elementLen * (remOff + remOffset)) ;

    int* block = blocks.addSome(blockLen) ;

    *(block++) = remId ;
    *(block++) = elementLen * vol ;
    *(block++) = numStrLevels ;
    *(block++) = elementLen * contigLen ;

    int* bexts  = block ;
    int* bstrs1 = bexts + numStrLevels ;
    int* bstrs2 = bstrs1 + numStrLevels ;
    for(int i = 0 ; i < numStrLevels ; i++) {
      *(bexts++)  = *(exts++) ;
      *(bstrs1++) = elementLen * *(locStrs++) ;
      *(bstrs2++) = elementLen * *(remStrs++) ;
    }

    numBlk++ ;
  }

  void** getLocs() {return locPtrs.data() ; }
  void** getRems() {return remPtrs.data() ; }
  int*   getBlks() {return blocks.data()  ; }

  const int rank, elementLen ;
  const int blockLen ;
  int numBlk ;

private :

  void* locBase ;
  int remOff ;
  void** remBases ;

  Vec<void*, 16> locPtrs, remPtrs ;
  Vec<int*,  16> blkPtrs ;

  Vec<int, 16> blocks ;
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//    The `CollBlockRMSchedule' class.  Collective version---adds      //
//                                      synchronization.               //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class CollBlockRMSchedule : public BlockRMSchedule {
public :
  CollBlockRMSchedule(void* _locBase, const int _remOff, void** _remBases,
                      const int _elementLen, const int _rank) :
          BlockRMSchedule(_locBase, _remOff, _remBases, _elementLen, _rank) {}

  void build(TreeSchedule* _spanTree) {spanTree = _spanTree ; }

  void gather() {
    spanTree->synch() ;
    BlockRMSchedule :: gather() ;
    spanTree->synch() ;
  }

  void scatter() {
    spanTree->synch() ;
    BlockRMSchedule :: scatter() ;
    ARMCI_AllFence() ;  // Wait for all locally-initiated Put's to
                        // complete remotely.
    spanTree->synch() ;
  }

//  void combine(void (*op)(void*, void*)) ;

private :
public :  // debug
  TreeSchedule* spanTree ;
} ;



// ----------------------------------------------------------------------
// Things to do
//

